@@ -392,21 +392,7 @@ def singleton_imputation_objective_function_of_all_guides(rep_i: int,
     return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
 
 
-
-class SingletonImputationOptimizer:
-    singleton_imputation_objective_function_of_each_guide_set = staticmethod(singleton_imputation_objective_function_of_each_guide_set)
-    singleton_imputation_objective_function_of_all_guides = staticmethod(singleton_imputation_objective_function_of_all_guides)
-
-    @staticmethod
-    def optimize_singleton_imputation_prior_strength(
-        experiment_guide_sets: ExperimentGuideSets, 
-        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
-        negative_control_guide_control_population_total_normalized_counts_reps: List[float], 
-        replicate_indices: List[int],
-        KL_guide_set_weights: Union[List[float], None],
-        cores: int) -> Tuple[List[float], List[float]]:
-
-        def weighted_objective_function_of_all_guides(rep_i: int,
+def singleton_imputation_weighted_objective_function_of_all_guides(rep_i: int,
                                                         experiment_guide_sets: ExperimentGuideSets,
                                                         negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
                                                         negative_control_guide_control_population_total_normalized_counts_reps: List[float],
@@ -420,12 +406,26 @@ class SingletonImputationOptimizer:
                 combined_score = (KL_guide_set_weights[0]*KL_guide_imputation_score_total_negative_avg) + (KL_guide_set_weights[1]*KL_guide_imputation_score_total_positive_avg) + (KL_guide_set_weights[2]*KL_guide_imputation_score_total_observation_avg)
                 return combined_score
 
+class SingletonImputationOptimizer:
+    singleton_imputation_objective_function_of_each_guide_set = staticmethod(singleton_imputation_objective_function_of_each_guide_set)
+    singleton_imputation_objective_function_of_all_guides = staticmethod(singleton_imputation_objective_function_of_all_guides)
+    singleton_imputation_weighted_objective_function_of_all_guides = staticmethod(singleton_imputation_weighted_objective_function_of_all_guides)
+
+    @staticmethod
+    def optimize_singleton_imputation_prior_strength(
+        experiment_guide_sets: ExperimentGuideSets, 
+        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+        negative_control_guide_control_population_total_normalized_counts_reps: List[float], 
+        replicate_indices: List[int],
+        KL_guide_set_weights: Union[List[float], None],
+        cores: int) -> Tuple[List[float], List[float]]:
+
 
 
 
         singleton_imputation_prior_strength_selected: List[float] = []
         for rep_i in replicate_indices:
-            weighted_objective_function_of_all_guides_p = functools.partial(weighted_objective_function_of_all_guides, rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, KL_guide_set_weights)
+            weighted_objective_function_of_all_guides_p = functools.partial(singleton_imputation_weighted_objective_function_of_all_guides, rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, KL_guide_set_weights)
             
             param_vals=[]
             loss_vals=[]
@@ -456,7 +456,7 @@ class SingletonImputationOptimizer:
             if res.success is True:
                 singleton_imputation_prior_strength = res.x
                 
-                KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = SingletonImputationOptimizer.objective_function_of_all_guides(rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, res.x)
+                KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = singleton_imputation_objective_function_of_all_guides(rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, res.x)
 
                 print("KL Negative Set Average: {}".format(KL_guide_imputation_score_total_negative_avg))
                 print("KL Positive Set Average: {}".format(KL_guide_imputation_score_total_positive_avg))
@@ -470,6 +470,8 @@ class SingletonImputationOptimizer:
         singleton_imputation_prior_strength_selected = np.asarray(singleton_imputation_prior_strength_selected).transpose().flatten()
         return singleton_imputation_prior_strength_selected
 
+
+
 def neighborhood_imputation_objective_function_of_each_guide_set(rep_i: int,
                                             guide_set: List[Guide],
                                             neighborhood_optimization_guide_sample_size: int,
@@ -544,9 +546,26 @@ def neighborhood_imputation_objective_function_of_all_guides(rep_i: int,
 
     return KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg
 
+def neighborhood_imputation_weighted_objective_function_of_all_guides(rep_i: int, 
+                                        experiment_guide_sets: ExperimentGuideSets,
+                                        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+                                        negative_control_guide_control_population_total_normalized_counts_reps: List[float],
+                                        neighborhood_bandwidth: float,
+                                        deviation_weights: List[float],
+                                        guide_set_weights: Union[List[float], None], 
+                                        params: Tuple[float, float]):
+    KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = NeighborhoodImputationOptimizer.neighborhood_imputation_objective_function_of_all_guides(rep_i, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
+    
+    if KL_guide_set_weights is None:
+        return KL_guide_imputation_score_total_combined_avg
+    else:
+        combined_score = (KL_guide_set_weights[0]*KL_guide_imputation_score_total_negative_avg) + (KL_guide_set_weights[1]*KL_guide_imputation_score_total_positive_avg) + (KL_guide_set_weights[2]*KL_guide_imputation_score_total_observation_avg)
+        return combined_score
+
 class NeighborhoodImputationOptimizer:
     neighborhood_imputation_objective_function_of_each_guide_set = staticmethod(neighborhood_imputation_objective_function_of_each_guide_set)
     neighborhood_imputation_objective_function_of_all_guides = staticmethod(neighborhood_imputation_objective_function_of_all_guides)
+    neighborhood_imputation_weighted_objective_function_of_all_guides = staticmethod(neighborhood_imputation_weighted_objective_function_of_all_guides)
 
     @staticmethod
     def optimize_neighborhood_imputation_prior_strength(
@@ -560,28 +579,11 @@ class NeighborhoodImputationOptimizer:
         cores: int,
         neighborhood_optimization_guide_sample_size: int = 50) -> Tuple[List[float], List[float]]:
 
-
-        def weighted_objective_function_of_all_guides(rep_i: int, 
-                                        experiment_guide_sets: ExperimentGuideSets,
-                                        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
-                                        negative_control_guide_control_population_total_normalized_counts_reps: List[float],
-                                        neighborhood_bandwidth: float,
-                                        deviation_weights: List[float],
-                                        guide_set_weights: Union[List[float], None], 
-                                        params: Tuple[float, float]):
-            KL_guide_imputation_score_total_combined_avg, KL_guide_imputation_score_total_negative_avg, KL_guide_imputation_score_total_positive_avg, KL_guide_imputation_score_total_observation_avg = NeighborhoodImputationOptimizer.neighborhood_imputation_objective_function_of_all_guides(rep_i, neighborhood_optimization_guide_sample_size, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, params)
-            
-            if KL_guide_set_weights is None:
-                return KL_guide_imputation_score_total_combined_avg
-            else:
-                combined_score = (KL_guide_set_weights[0]*KL_guide_imputation_score_total_negative_avg) + (KL_guide_set_weights[1]*KL_guide_imputation_score_total_positive_avg) + (KL_guide_set_weights[2]*KL_guide_imputation_score_total_observation_avg)
-                return combined_score
-
         neighborhood_imputation_prior_strength_selected: List[float] = []
         neighborhood_imputation_likelihood_strength_selected: List[float] = []
 
         for rep_i in replicate_indices:
-            weighted_objective_function_of_all_guides_p = functools.partial(weighted_objective_function_of_all_guides, rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, KL_guide_set_weights)
+            weighted_objective_function_of_all_guides_p = functools.partial(neighborhood_imputation_weighted_objective_function_of_all_guides, rep_i, experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_bandwidth, deviation_weights, KL_guide_set_weights)
 
 
             param_vals=[]
@@ -638,66 +640,43 @@ class NeighborhoodImputationOptimizer:
 
         return neighborhood_imputation_prior_strength_selected, neighborhood_imputation_likelihood_strength_selected
 
-class ShrinkageOptimizer:
+
+
+@staticmethod
+def shrinkage_optimizer_weighted_objective_function_of_all_guides(rep_i: int, 
+                                        spatial_guides_for_fit: List[Guide],
+                                        singleton_guides_for_fit: List[Guide], 
+                                        neighborhood_experiment_guide_sets: ExperimentGuideSets,
+                                        negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
+                                        negative_control_guide_control_population_total_normalized_counts_reps: List[float],
+                                        neighborhood_imputation_prior_strength: List[float],
+                                        neighborhood_imputation_likelihood_strength: List[float],
+                                        singleton_imputation_prior_strength: List[float],
+                                        neighborhood_bandwidth: float,
+                                        enable_neighborhood_prior: bool,
+                                        monte_carlo_trials: int,
+                                        random_seed: Union[int, None],
+                                        params: Tuple[float]):
+    # TODO: There is a major inefficiency for this optimization function and the other optimization function, is that even though optimization is done for each rep_i, the posterior is calculated for all reps then indexed for the rep_i argument... I think being able to index the arguments to the perform_score_imputation should work if the typehint is correct. 
+    shrinkage_prior_strength_test=params[0]
     
-    @staticmethod
-    def weighted_objective_function_of_all_guides(rep_i: int, 
-                                            spatial_guides_for_fit: List[Guide],
-                                            singleton_guides_for_fit: List[Guide], 
-                                            neighborhood_experiment_guide_sets: ExperimentGuideSets,
-                                            negative_control_guide_sample_population_total_normalized_counts_reps: List[float],
-                                            negative_control_guide_control_population_total_normalized_counts_reps: List[float],
-                                            neighborhood_imputation_prior_strength: List[float],
-                                            neighborhood_imputation_likelihood_strength: List[float],
-                                            singleton_imputation_prior_strength: List[float],
-                                            neighborhood_bandwidth: float,
-                                            enable_neighborhood_prior: bool,
-                                            monte_carlo_trials: int,
-                                            random_seed: Union[int, None],
-                                            params: Tuple[float]):
-        # TODO: There is a major inefficiency for this optimization function and the other optimization function, is that even though optimization is done for each rep_i, the posterior is calculated for all reps then indexed for the rep_i argument... I think being able to index the arguments to the perform_score_imputation should work if the typehint is correct. 
-        shrinkage_prior_strength_test=params[0]
-        
-        total_normalized_counts_per_guide_spatial : List[float] = [each_guide.sample_population_normalized_count_reps[rep_i] + each_guide.control_population_normalized_count_reps[rep_i] for each_guide in spatial_guides_for_fit]
-        total_normalized_counts_per_guide_singleton : List[float] = [each_guide.sample_population_normalized_count_reps[rep_i] + each_guide.control_population_normalized_count_reps[rep_i] for each_guide in singleton_guides_for_fit]
-        
-        # NOTE: The first list corresponds to each guide, the second list corresponds to number of replicates, the value is the mean LFC
-        guide_count_posterior_LFC_normalized_mean_list_per_guide_spatial : List[float] = []
-        guide_count_posterior_LFC_normalized_mean_list_per_guide_singleton : List[float] = []
-
-        if enable_neighborhood_prior:
-            for each_guide in spatial_guides_for_fit:
-                # TODO: The code for calculating the posterior inputs for the spatial_imputation model could be modularized so that there are not any repetitive code
-
-                # By default, set the unweighted prior as the negative control normalized counts
-                unweighted_prior_alpha: float = np.asarray([negative_control_guide_sample_population_total_normalized_counts_reps[rep_i]])
-                unweighted_prior_beta: float = np.asarray([negative_control_guide_control_population_total_normalized_counts_reps[rep_i]])
-
-                # If able to use spatial information, replace the unweighted priors with the spatial imputational posterior
-                imputation_posterior_alpha, imputation_posterior_beta, _, _ = ModelInference.perform_neighboorhood_score_imputation(each_guide, neighborhood_experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_imputation_prior_strength, neighborhood_imputation_likelihood_strength, [rep_i], neighborhood_bandwidth)
-
-                # Propogate the imputation posterior to the shrinkage prior
-                unweighted_prior_alpha = imputation_posterior_alpha
-                unweighted_prior_beta = imputation_posterior_beta
-
-                shrinkage_result: ShrinkageResult = ModelInference.perform_score_shrinkage(each_guide, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, shrinkage_prior_strength_test, unweighted_prior_alpha, unweighted_prior_beta, monte_carlo_trials, random_seed, [rep_i])
-
-                # NOTE: List[List[float]], first list is each replicate, second list is the monte-carlo samples. We want the mean of the monte-carlo samples next
-                guide_count_posterior_LFC_samples_normalized_list: List[List[float]] = shrinkage_result.guide_count_posterior_LFC_samples_normalized_list
-                
-                # This corresponds to the guide posterior mean LFC for each replicate separately for shrinkage prior weight optimization. After optimization of the shrinkage weight, the mean LFC of the averaged posterior of the replicates will be used.
-                guide_count_posterior_LFC_normalized_mean_list: List[float] = np.mean(guide_count_posterior_LFC_samples_normalized_list, axis=0) 
-                guide_count_posterior_LFC_normalized_mean_list_per_guide_spatial.append(guide_count_posterior_LFC_normalized_mean_list[rep_i])
-        
-        for each_guide in singleton_guides_for_fit:
+    total_normalized_counts_per_guide_spatial : List[float] = [each_guide.sample_population_normalized_count_reps[rep_i] + each_guide.control_population_normalized_count_reps[rep_i] for each_guide in spatial_guides_for_fit]
+    total_normalized_counts_per_guide_singleton : List[float] = [each_guide.sample_population_normalized_count_reps[rep_i] + each_guide.control_population_normalized_count_reps[rep_i] for each_guide in singleton_guides_for_fit]
+    
+    # NOTE: The first list corresponds to each guide, the second list corresponds to number of replicates, the value is the mean LFC
+    guide_count_posterior_LFC_normalized_mean_list_per_guide_spatial : List[float] = []
+    guide_count_posterior_LFC_normalized_mean_list_per_guide_singleton : List[float] = []
 
+    if enable_neighborhood_prior:
+        for each_guide in spatial_guides_for_fit:
             # TODO: The code for calculating the posterior inputs for the spatial_imputation model could be modularized so that there are not any repetitive code
-            imputation_posterior_alpha, imputation_posterior_beta = ModelInference.perform_singleton_score_imputation(each_guide, 
-                    negative_control_guide_sample_population_total_normalized_counts_reps, 
-                    negative_control_guide_control_population_total_normalized_counts_reps, 
-                    singleton_imputation_prior_strength,
-                    [rep_i])
 
+            # By default, set the unweighted prior as the negative control normalized counts
+            unweighted_prior_alpha: float = np.asarray([negative_control_guide_sample_population_total_normalized_counts_reps[rep_i]])
+            unweighted_prior_beta: float = np.asarray([negative_control_guide_control_population_total_normalized_counts_reps[rep_i]])
+
+            # If able to use spatial information, replace the unweighted priors with the spatial imputational posterior
+            imputation_posterior_alpha, imputation_posterior_beta, _, _ = ModelInference.perform_neighboorhood_score_imputation(each_guide, neighborhood_experiment_guide_sets, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, neighborhood_imputation_prior_strength, neighborhood_imputation_likelihood_strength, [rep_i], neighborhood_bandwidth)
 
             # Propogate the imputation posterior to the shrinkage prior
             unweighted_prior_alpha = imputation_posterior_alpha
@@ -705,27 +684,54 @@ class ShrinkageOptimizer:
 
             shrinkage_result: ShrinkageResult = ModelInference.perform_score_shrinkage(each_guide, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, shrinkage_prior_strength_test, unweighted_prior_alpha, unweighted_prior_beta, monte_carlo_trials, random_seed, [rep_i])
 
-
             # NOTE: List[List[float]], first list is each replicate, second list is the monte-carlo samples. We want the mean of the monte-carlo samples next
             guide_count_posterior_LFC_samples_normalized_list: List[List[float]] = shrinkage_result.guide_count_posterior_LFC_samples_normalized_list
             
             # This corresponds to the guide posterior mean LFC for each replicate separately for shrinkage prior weight optimization. After optimization of the shrinkage weight, the mean LFC of the averaged posterior of the replicates will be used.
             guide_count_posterior_LFC_normalized_mean_list: List[float] = np.mean(guide_count_posterior_LFC_samples_normalized_list, axis=0) 
+            guide_count_posterior_LFC_normalized_mean_list_per_guide_spatial.append(guide_count_posterior_LFC_normalized_mean_list[rep_i])
+    
+    for each_guide in singleton_guides_for_fit:
 
-            guide_count_posterior_LFC_normalized_mean_list_per_guide_singleton.append(guide_count_posterior_LFC_normalized_mean_list[rep_i]) 
+        # TODO: The code for calculating the posterior inputs for the spatial_imputation model could be modularized so that there are not any repetitive code
+        imputation_posterior_alpha, imputation_posterior_beta = ModelInference.perform_singleton_score_imputation(each_guide, 
+                negative_control_guide_sample_population_total_normalized_counts_reps, 
+                negative_control_guide_control_population_total_normalized_counts_reps, 
+                singleton_imputation_prior_strength,
+                [rep_i])
 
-        #
-        # Calculate the Breusch-Pagan statistic
-        #
-        # Prepare X - which is the normalized count, since the objective is to reduce hederoscedasticity across count
-        total_normalized_count_per_guide_X: List[float] = np.concatenate([total_normalized_counts_per_guide_spatial, total_normalized_counts_per_guide_singleton])
 
-        # Prepare Y - which is the LFC score, since we want to reduce heteroscedasticity of the LFC
-        LFC_posterior_mean_per_guide_Y: List[float] = np.concatenate([guide_count_posterior_LFC_normalized_mean_list_per_guide_spatial, guide_count_posterior_LFC_normalized_mean_list_per_guide_singleton])
+        # Propogate the imputation posterior to the shrinkage prior
+        unweighted_prior_alpha = imputation_posterior_alpha
+        unweighted_prior_beta = imputation_posterior_beta
 
-        BP_statistic = StatisticalHelperMethods.calculate_breusch_pagan(total_normalized_count_per_guide_X, LFC_posterior_mean_per_guide_Y) 
+        shrinkage_result: ShrinkageResult = ModelInference.perform_score_shrinkage(each_guide, negative_control_guide_sample_population_total_normalized_counts_reps, negative_control_guide_control_population_total_normalized_counts_reps, shrinkage_prior_strength_test, unweighted_prior_alpha, unweighted_prior_beta, monte_carlo_trials, random_seed, [rep_i])
+
+
+        # NOTE: List[List[float]], first list is each replicate, second list is the monte-carlo samples. We want the mean of the monte-carlo samples next
+        guide_count_posterior_LFC_samples_normalized_list: List[List[float]] = shrinkage_result.guide_count_posterior_LFC_samples_normalized_list
         
-        return BP_statistic
+        # This corresponds to the guide posterior mean LFC for each replicate separately for shrinkage prior weight optimization. After optimization of the shrinkage weight, the mean LFC of the averaged posterior of the replicates will be used.
+        guide_count_posterior_LFC_normalized_mean_list: List[float] = np.mean(guide_count_posterior_LFC_samples_normalized_list, axis=0) 
+
+        guide_count_posterior_LFC_normalized_mean_list_per_guide_singleton.append(guide_count_posterior_LFC_normalized_mean_list[rep_i]) 
+
+    #
+    # Calculate the Breusch-Pagan statistic
+    #
+    # Prepare X - which is the normalized count, since the objective is to reduce hederoscedasticity across count
+    total_normalized_count_per_guide_X: List[float] = np.concatenate([total_normalized_counts_per_guide_spatial, total_normalized_counts_per_guide_singleton])
+
+    # Prepare Y - which is the LFC score, since we want to reduce heteroscedasticity of the LFC
+    LFC_posterior_mean_per_guide_Y: List[float] = np.concatenate([guide_count_posterior_LFC_normalized_mean_list_per_guide_spatial, guide_count_posterior_LFC_normalized_mean_list_per_guide_singleton])
+
+    BP_statistic = StatisticalHelperMethods.calculate_breusch_pagan(total_normalized_count_per_guide_X, LFC_posterior_mean_per_guide_Y) 
+    
+    return BP_statistic
+
+
+class ShrinkageOptimizer:
+    shrinkage_optimizer_weighted_objective_function_of_all_guides = staticmethod(shrinkage_optimizer_weighted_objective_function_of_all_guides)
 
     @staticmethod
     def optimize_shrinkage_prior_strength(neighborhood_experiment_guide_sets: Union[ExperimentGuideSets, None],
@@ -769,7 +775,7 @@ class ShrinkageOptimizer:
 
         shrinkage_prior_strength_selected: List[float] = []
         for rep_i in replicate_indices:
-            weighted_objective_function_of_all_guides_p = functools.partial(ShrinkageOptimizer.weighted_objective_function_of_all_guides, rep_i,
+            weighted_objective_function_of_all_guides_p = functools.partial(shrinkage_optimizer_weighted_objective_function_of_all_guides, rep_i,
                                             spatial_guides_for_fit,
                                             singleton_guides_for_fit,
                                             neighborhood_experiment_guide_sets,
